from typing import Optional, List

from django.conf import settings
from django.utils.functional import SimpleLazyObject

from rhazes.dependency import DependencyResolver
from rhazes.protocol import BeanProtocol
from rhazes.scanner import ModuleScanner, class_scanner


class ApplicationContext:
    _initialized = False
    _builder_registry = {}
    _additional_roots_to_scan = []

    @classmethod
    def get_roots_to_scan(cls):
        roots_to_scan = cls._roots_to_scan()
        return cls._additional_roots_to_scan + roots_to_scan

    @classmethod
    def _initialize_beans(cls):
        beans = set()
        modules = ModuleScanner(cls.get_roots_to_scan()).scan()
        for module in modules:
            scanned_classes = class_scanner(module)
            for scanned_class in scanned_classes:
                if issubclass(scanned_class, (BeanProtocol,)):
                    beans.add(scanned_class)

        for clazz, obj in DependencyResolver(beans).resolve().items():
            cls.register_bean(clazz, obj)

    @classmethod
    def _roots_to_scan(cls):
        if hasattr(settings, "RHAZES_PACKAGES"):
            return settings.RHAZES_PACKAGES
        return settings.INSTALLED_APPS

    @classmethod
    def initialize(cls, additional_roots_to_scan: List[str] = None):
        if additional_roots_to_scan is not None:
            cls._additional_roots_to_scan = additional_roots_to_scan

        if cls._initialized:
            return
        cls._initialize_beans()
        cls._initialized = True

    @classmethod
    def register_bean(cls, clazz, builder, override=False):
        if clazz not in cls._builder_registry or override:
            cls._builder_registry[clazz] = builder

    @classmethod
    def get_bean(cls, of: type) -> Optional:
        builder = cls._builder_registry.get(of)
        if builder is None:
            return None
        return builder(cls)

    @classmethod
    def get_lazy_bean(cls, of: type) -> Optional:
        return SimpleLazyObject(lambda: cls.get_bean(of))



from typing import Set, Type

from rhazes.collections.stack import UniqueStack
from rhazes.exceptions import DependencyCycleException
from rhazes.protocol import (
    BeanProtocol,
    BeanFactory,
    DependencyNodeMetadata,
    DependencyNode,
)


class NodeBuilder:
    def __init__(self, node: DependencyNode, metadata: DependencyNodeMetadata):
        self.node = node
        self.metadata = metadata

    def build(self, application_context):
        args: list = self.metadata.args
        dependency_positions = self.metadata.dependency_position
        for dep in self.metadata.dependencies:
            lazy = (
                self.metadata.lazy_dependencies is not None
                and dep in self.metadata.lazy_dependencies
            )
            args[dependency_positions[dep]] = (
                application_context.get_lazy_bean(dep)
                if lazy
                else application_context.get_bean(dep)
            )
        if self.metadata.is_factory:
            factory: BeanFactory = self.node.cls(*self.metadata.args)
            return factory.produce()
        else:
            return self.node.cls(*self.metadata.args)


class SingletonNodeBuilder:
    def __init__(self, builder: NodeBuilder):
        self.builder = builder
        self.instance = None

    def build(self, application_context):
        if self.instance is not None:
            return self.instance
        self.instance = self.builder.build(application_context)
        return self.instance


class DependencyResolver:
    def __init__(self, bean_classes: Set[Type[BeanProtocol]] = None):
        self.bean_classes: Set[Type[BeanProtocol]] = (
            bean_classes if bean_classes is not None else set()
        )
        self.bean_interface_map = {}
        self.fill_bean_interface_map()
        self.objects = {}
        self.node_registry = {}
        self.node_metadata_registry = {}

    def __add_or_init_interface_mapping(self, interface, cls):
        mapping = self.bean_interface_map.get(interface, [])
        mapping.append(cls)
        self.bean_interface_map[interface] = mapping

    def fill_bean_interface_map(self):
        # Map list of implementations (value) for each bean interface (value)
        for bean_class in self.bean_classes:
            if bean_class.bean_details().bean_for is not None:
                self.__add_or_init_interface_mapping(
                    bean_class.bean_details().bean_for, bean_class
                )

        # Find primary implementation of each interface and update the map
        for bean_for, implementations in self.bean_interface_map.items():
            primary = implementations[0]
            if len(implementations) > 1:
                for implementation in implementations:
                    if implementation.bean_details().primary:
                        primary = implementation
                        break
            self.bean_interface_map[bean_for] = primary

    def register_dependency_node(self, cls) -> DependencyNode:

        # If the dependency is not a bean (it could be the interface of another bean) then look for implementation bean
        if cls in self.bean_interface_map:
            return self.register_dependency_node(self.bean_interface_map[cls])

        # If we have already created DependencyNode for this class then return it
        if cls in self.node_registry:
            return self.node_registry[cls]
        node = DependencyNode(cls)
        self.node_registry[cls] = node

        # If the bean is presenting an interface then register it for that interface too
        if issubclass(cls, (BeanProtocol,)) and cls.bean_details().bean_for is not None:
            self.node_registry[cls.bean_details().bean_for] = node

        return node

    def register_metadata(self, cls) -> DependencyNodeMetadata:
        """
        Creates DependencyNodeMetadata and registers it for the cls in self.node_metadata_registry
        :param cls: class to generate DependencyNodeMetadata for
        :return generated DependencyNodeMetadata
        """
        metadata = DependencyNodeMetadata.generate(
            cls, self.bean_classes, self.bean_interface_map
        )
        self.node_metadata_registry[cls] = metadata
        return metadata

    def resolve(self) -> dict:
        """
        Resolves dependencies by building graph, traverse it and returning list of created objects
        :returns dictionary of created objects for each class
        """

        to_process = []

        # Building Graph
        for cls in self.bean_classes:
            metadata = self.register_metadata(cls)
            node = self.register_dependency_node(cls)
            for dependency in metadata.dependencies:
                node.add_dependency(self.register_dependency_node(dependency))
            to_process.append(node)

        # Processing each node
        for node in to_process:
            self._process(node, UniqueStack())

        return self.objects

    def _process(self, node, stack):
        """
        Depth first traversal on nodes.
            Base case: node is already built, so we ignore building again
            Using "post-order" BFS to use a UniqueStack in order to detect cycles

        Accepts any node and if its already processed ignores it.
        The reason is that we aren't sure we have a single dependency tree or graph or multiple ones
        That's why this method is called for all bean classes

        :param node: a node to start processing from
        :param stack: instance of UniqueStack for dependency cycle detection
        :return:
        """
        if node.cls in self.objects:
            return
        try:
            stack.append(node)
        except ValueError:  # item is not unique in the stack
            raise DependencyCycleException(stack, node)
        for child in node.dependencies:
            self._process(child, stack)
        self.generate_builder(node)
        stack.pop()

    def generate_builder(self, node: DependencyNode):
        metadata: DependencyNodeMetadata = self.node_metadata_registry[node.cls]

        if metadata.is_factory:
            clazz = metadata.bean_for
        else:
            clazz = node.cls

        node_builder = NodeBuilder(node, metadata)
        if metadata.is_singleton:
            node_builder = SingletonNodeBuilder(node_builder)

        self.objects[clazz] = node_builder.build
        if metadata.bean_for is not None and (
            metadata.is_factory
            or node.cls == self.bean_interface_map[metadata.bean_for]
        ):
            self.objects[metadata.bean_for] = node_builder.build



from rhazes.context import ApplicationContext


class TemporaryContext:
    def __init__(self):
        self.old_values = {}

    def register_bean(self, clazz, obj):
        old_value = ApplicationContext.get_bean(clazz)
        if clazz not in self.old_values:
            self.old_values[clazz] = old_value
        ApplicationContext.register_bean(clazz, lambda x: obj, True)

    def reset(self):
        for clazz, old_value in self.old_values.items():
            ApplicationContext.register_bean(clazz, lambda x: old_value, True)


class TemporaryContextManager:
    def __init__(self):
        self.temporary_context = TemporaryContext()

    def __enter__(self):
        return self.temporary_context

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.temporary_context.reset()



import inspect
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pydoc import locate
from typing import (
    Protocol,
    runtime_checkable,
    Optional,
    Iterable,
    Dict,
    Type,
    TypedDict,
    List,
)

from rhazes.exceptions import MissingDependencyException

logger = logging.getLogger(__name__)


@dataclass(unsafe_hash=True)
class BeanDetails:
    bean_for: Optional[type]
    primary: bool = False
    singleton: bool = False
    lazy_dependencies: Optional[List[type]] = None


@runtime_checkable
class BeanProtocol(Protocol):
    @classmethod
    def bean_details(cls) -> BeanDetails:
        ...


class BeanFactory(ABC):
    @classmethod
    def produces(cls):
        """
        Determines what type of object (class) this factory produces
        :return: class of which the object in factory method will be
        """
        if issubclass(cls, (BeanProtocol,)) and cls.bean_details().bean_for is not None:
            return cls.bean_details().bean_for
        raise NotImplementedError

    @abstractmethod
    def produce(self):
        """
        :return: object of type from cls.produces()
        """
        pass


@dataclass
class DependencyNodeMetadata:
    """
    - dependencies: list of dependency classes
    - dependency_position: dictionary of dependency class positions in arguments
    - args: list of prefilled arguments to be used as *args for constructing
    """

    dependencies: Optional[list]
    dependency_position: Optional[dict]
    args: Optional[list]
    bean_for: Optional[Type] = None
    is_factory: bool = False
    is_singleton: bool = False
    lazy_dependencies: Optional[list] = None

    @staticmethod
    def generate(
        cls,
        bean_classes: Iterable[BeanProtocol],
        bean_interface_mapping: Dict[Type, Type],
    ):
        """
        Generates DependencyNodeMetadata instance for a class (cls) after validating its constructor dependencies
        :param cls: class to generate DependencyNodeMetadata for
        :param bean_classes: other bean classes, possible to depend on
        :param bean_interface_mapping: possible classes to depend on
        :return: generated DependencyNodeMetadata
        """
        args = []
        dependencies = []
        dependency_position = {}
        signature = inspect.signature(cls.__init__)
        i = 0
        for k, v in signature.parameters.items():
            if k == "self":
                continue

            if k in ["args", "kwargs"]:
                logger.warning(
                    f"bean class {cls} has __init__ which uses *args or **kwargs. "
                    f"It's impossible to detect the inputs"
                )
                continue

            clazz = None

            if type(v.annotation) == str:
                clazz = locate(f"{cls.__module__}.{v.annotation}")
                if clazz is None:
                    raise Exception(f"Failed to locate {v.annotation}")
            else:
                clazz = v.annotation

            if clazz in bean_classes:
                dependencies.append(clazz)
                args.append(None)
                dependency_position[clazz] = i
            elif clazz in bean_interface_mapping:
                dependencies.append(bean_interface_mapping[clazz])
                args.append(None)
                dependency_position[bean_interface_mapping[clazz]] = i
            elif v.default == v.empty:
                raise MissingDependencyException(cls, clazz)
            else:
                args.append(v.default)
            i += 1

        bean_for = None
        is_factory = False
        is_singleton = False
        lazy_dependencies = None

        if issubclass(cls, (BeanProtocol,)):
            bean_for = cls.bean_details().bean_for
            is_singleton = cls.bean_details().singleton
            lazy_dependencies = cls.bean_details().lazy_dependencies
        if issubclass(cls, (BeanFactory,)):
            bean_for = bean_for if bean_for is not None else cls.produces()
            is_factory = True

        return DependencyNodeMetadata(
            dependencies,
            dependency_position,
            args,
            bean_for,
            is_factory,
            is_singleton,
            lazy_dependencies,
        )


class DependencyNode:
    def __init__(self, cls):
        self.cls = cls
        self.dependencies = []

    def add_dependency(self, dependency: "DependencyNode"):
        self.dependencies.append(dependency)

    def __str__(self):
        return str(self.cls)


class InjectionConfiguration(TypedDict):
    lazy: bool



class MissingDependencyException(Exception):
    def __init__(self, cls, missing):
        self.cls = cls
        self.missing = missing
        super().__init__(
            f"Class {cls} depends on {missing} which was not found during scan!"
        )


class DependencyCycleException(Exception):
    def __init__(self, stack, breaker):
        cycle = " -> ".join([str(item) for item in stack])
        cycle += f" -> {str(breaker)}"
        self.stack = stack
        self.breaker = breaker
        super().__init__(f"Dependency Cycle Detected!\n{cycle}")



from collections import deque


class UniqueStack(deque):
    def _validate_unique(self, value):
        for i in self:
            if value == i:
                raise ValueError("Item already in stack")

    def append(self, *args, **kwargs):
        self._validate_unique(args[0])
        return super(UniqueStack, self).append(*args, **kwargs)

    def appendleft(self, *args, **kwargs) -> None:
        self._validate_unique(args[0])
        return super(UniqueStack, self).appendleft(*args, **kwargs)



import importlib
import inspect
from typing import List, Set


def class_scanner(module: str):
    result = []
    for _, cls in inspect.getmembers(importlib.import_module(module), inspect.isclass):
        if cls.__module__ == module:
            result.append(cls)
    return result


class ModuleScanner:
    """
    Scans for python packages submodules recursively
    """

    def __init__(self, roots_to_scan: List[str]):
        self.roots_to_scan = roots_to_scan

    def _list_submodules(self, module) -> list[str]:
        """
        Args:
            module: The module to list submodules from.
        """
        # We first respect __all__ attribute if it already defined.
        submodules = getattr(module, "__all__", None)
        if submodules:
            return submodules

        # Then, we respect __init__.py file to get imported submodules.
        import inspect

        submodules = [
            o[0] for o in inspect.getmembers(module) if inspect.ismodule(o[1])
        ]
        if submodules:
            return submodules

        # Finally we can just scan for submodules via pkgutil.
        import pkgutil

        # pkgutill will invoke `importlib.machinery.all_suffixes()` to
        # determine whether a file is a module or not, so if you get
        # any submoudles that are unexpected to get, you need to check
        # this function to do the confirmation.
        # If you want to retrive a directory as a submoudle, you will
        # need to clarify this by putting a `__init__.py`` file in the
        # folder, even for Python3.
        if hasattr(module, "__path__"):
            return [x.name for x in pkgutil.iter_modules(module.__path__)]
        return []

    def scan(self) -> Set[str]:
        packages = set()
        for pkg in self.roots_to_scan:
            packages.update(self._scan(pkg))
        return packages

    def _scan(self, pkg) -> Set[str]:
        packages = set()
        try:
            mod = importlib.import_module(pkg)
            for item in self._list_submodules(mod):
                sub_m = f"{pkg}.{item}"
                packages.update(self._scan(sub_m))
            packages.add(pkg)
        except ModuleNotFoundError:
            pass
        return packages



import functools
import inspect
from typing import Optional, Dict, Type, List, Union

from rhazes.context import ApplicationContext
from rhazes.protocol import InjectionConfiguration, BeanDetails


def bean(
    _for=None,
    primary=False,
    singleton=False,
    lazy_dependencies: Optional[List[Union[type, str]]] = None,
):
    def decorator(cls):
        if _for is not None and not issubclass(cls, _for):
            raise Exception(
                f"{cls} bean is meant to be registered for interface {_for} "
                f"but its not a subclass of that interface"
            )

        @functools.wraps(cls, updated=())
        class DecoratedBean(cls):
            @classmethod
            def bean_details(cls) -> BeanDetails:
                return BeanDetails(_for, primary, singleton, lazy_dependencies)

        return DecoratedBean

    return decorator


def inject_kwargs(injections, configuration, func, kwargs: dict):
    signature = inspect.signature(func)
    for k, v in signature.parameters.items():
        if k in ["self", "cls"]:
            continue

        if k in ["args", "kwargs"]:
            continue

        if (injections is not None and v.annotation in injections) or (
            injections is None and v.annotation is not None
        ):
            lazy = configuration.get(v.annotation, {}).get("lazy", False)
            if lazy:
                dep = ApplicationContext.get_lazy_bean(v.annotation)
            else:
                dep = ApplicationContext.get_bean(v.annotation)
            if dep is not None:
                kwargs[k] = dep
        return kwargs


def inject(injections=None, configuration: Dict[Type, InjectionConfiguration] = None):
    if configuration is None:
        configuration = {}

    def decorator(obj_of_func):

        if isinstance(obj_of_func, type):
            # We are dealing with a class
            @functools.wraps(obj_of_func, updated=())
            class Proxy(obj_of_func):
                def __init__(self, *args, **kwargs):
                    inject_kwargs(
                        injections, configuration, obj_of_func.__init__, kwargs
                    )
                    super(obj_of_func, self).__init__(*args, **kwargs)

            return Proxy

        elif callable(obj_of_func):

            if obj_of_func.__name__ == "__init__":  # constructor

                def proxy(obj, *args, **kwargs):
                    inject_kwargs(injections, configuration, obj_of_func, kwargs)
                    return obj_of_func(obj, *args, **kwargs)

                return proxy
            else:

                def proxy(*args, **kwargs):
                    inject_kwargs(injections, configuration, obj_of_func, kwargs)
                    return obj_of_func(*args, **kwargs)

                return proxy

        else:
            return obj_of_func

    return decorator
